import struct
from typing import Any

from . import DeviceField, FieldName


class SwitchField(DeviceField):
    def __init__(self, name: FieldName, address: int):
        super().__init__(name, address, 1)

    def parse(self, data: bytes) -> bool:
        return struct.unpack("!H", data)[0] == 1

    def is_writeable(self):
        return True

    def allowed_write_type(self, value: Any) -> bool:
        return isinstance(value, bool)



import asyncio
import logging
import async_timeout
from typing import Any, Callable, List, cast
from bleak import BleakScanner
from bleak.exc import BleakError
from bleak_retry_connector import BleakClientWithServiceCache, establish_connection

from .encryption import BluettiEncryption, Message, MessageType
from ..registers import ReadableRegisters
from ..const import NOTIFY_UUID, WRITE_UUID
from ..base_devices import BluettiDevice

_LOGGER = logging.getLogger(__name__)


class DeviceReaderConfig:
    def __init__(self, timeout: int = 60, use_encryption: bool = False):
        self.timeout = timeout
        self.use_encryption = use_encryption


class DeviceReader:
    def __init__(
        self,
        mac: str,
        bluetti_device: BluettiDevice,
        future_builder_method: Callable[[], asyncio.Future[Any]],
        config: DeviceReaderConfig = DeviceReaderConfig(),
        lock: asyncio.Lock = asyncio.Lock(),
    ):
        self.mac = mac
        self.bluetti_device = bluetti_device
        self.create_future = future_builder_method
        self.config = config
        self.polling_lock = lock

        self.device = None
        self.client = None

        self.has_notifier = False
        self.current_registers = None
        self.notify_response = bytearray()
        self.notify_future: asyncio.Future[Any] | None = None
        self.encryption = BluettiEncryption()

    async def read(
        self, only_registers: List[ReadableRegisters] | None = None, raw: bool = False
    ) -> dict | None:

        registers = self.bluetti_device.get_polling_registers()

        if only_registers is not None:
            registers = only_registers

        parsed_data: dict = {}

        _LOGGER.debug("Reading device registers")

        async with self.polling_lock:
            try:
                async with async_timeout.timeout(self.config.timeout):
                    _LOGGER.debug("Searching for %s", self.mac)
                    self.device = await BleakScanner.find_device_by_address(
                        self.mac, timeout=5
                    )

                    if self.device is None:
                        _LOGGER.error("Device not found")
                        return

                    _LOGGER.debug("Connecting to %s", self.mac)

                    self.client = await establish_connection(
                        BleakClientWithServiceCache,
                        self.device,
                        self.device.name or "Unknown Device",
                        max_attempts=10,
                    )

                    _LOGGER.debug("Connected to device")

                    if not self.has_notifier:
                        await self.client.start_notify(
                            NOTIFY_UUID, self._notification_handler
                        )
                        self.has_notifier = True

                    _LOGGER.debug("Notification handler setup complete")

                    while (
                        self.config.use_encryption
                        and not self.encryption.is_ready_for_commands
                    ):
                        await asyncio.sleep(5)
                        _LOGGER.debug("Encryption handshake not finished yet")

                    for register in registers:
                        body = register.parse_response(
                            await self._async_send_command(register)
                        )

                        _LOGGER.debug("Raw data: %s", body)

                        if raw:
                            d = {}
                            d[register.starting_address] = body
                            parsed_data.update(d)
                            continue

                        parsed = self.bluetti_device.parse(
                            register.starting_address, body
                        )

                        _LOGGER.debug("Parsed data: %s", parsed)

                        parsed_data.update(parsed)

            except TimeoutError:
                _LOGGER.warning("Timeout")
                return None
            except BleakError as err:
                _LOGGER.warning("Bleak error: %s", err)
                return None
            except BaseException as err:
                _LOGGER.warning("Unknown error %s", err)
                return None
            finally:
                if self.has_notifier:
                    try:
                        await self.client.stop_notify(NOTIFY_UUID)
                        _LOGGER.debug("Stopped notifier")
                    except:
                        # Ignore errors here
                        pass
                    self.has_notifier = False
                if self.client:
                    await self.client.disconnect()
                    _LOGGER.debug("Disconnected from device")

            # Reset Encryption keys
            self.encryption.reset()

            # Check if dict is empty
            if not parsed_data:
                return None

            return parsed_data

    async def _async_send_command(self, registers: ReadableRegisters) -> bytes:
        """Send command and return response"""
        self.current_registers = registers
        self.notify_response = bytearray()
        self.notify_future = self.create_future()

        command_bytes = bytes(registers)

        # Encrypt command
        if self.config.use_encryption is True:
            if not self.encryption.is_ready_for_commands:
                return bytes()
            command_bytes = self.encryption.aes_encrypt(
                command_bytes, self.encryption.secure_aes_key, None
            )

        try:
            # Make request
            await self.client.write_gatt_char(WRITE_UUID, command_bytes)

            _LOGGER.debug("Request sent (%s)", registers)

            # Wait for response
            res = await asyncio.wait_for(self.notify_future, timeout=5)

            _LOGGER.debug("Got response")

            return cast(bytes, res)
        except:
            _LOGGER.warning("Error while reading data")

        return bytes()

    async def _notification_handler(self, _: int, data: bytearray):
        """Handle bt data."""
        _LOGGER.debug("Got new data")

        if self.config.use_encryption is True:
            message = Message(data)

            if message.is_pre_key_exchange:
                message.verify_checksum()

                if message.type == MessageType.CHALLENGE:
                    challenge_response = self.encryption.msg_challenge(message)
                    await self.client.write_gatt_char(WRITE_UUID, challenge_response)
                    return

                if message.type == MessageType.CHALLENGE_ACCEPTED:
                    _LOGGER.debug("Challenge accepted")
                    return

            if self.encryption.unsecure_aes_key is None:
                _LOGGER.error("Received encrypted message before key initialization")

            key, iv = self.encryption.getKeyIv()
            decrypted = Message(self.encryption.aes_decrypt(message.buffer, key, iv))

            if decrypted.is_pre_key_exchange:
                decrypted.verify_checksum()

                if decrypted.type == MessageType.PEER_PUBKEY:
                    peer_pubkey_response = self.encryption.msg_peer_pubkey(decrypted)
                    await self.client.write_gatt_char(WRITE_UUID, peer_pubkey_response)
                    return

                if decrypted.type == MessageType.PUBKEY_ACCEPTED:
                    self.encryption.msg_key_accepted(decrypted)
                    return

            # Handle as message
            data = decrypted.buffer

        # Save data
        self.notify_response.extend(data)
        self.notify_future.set_result(self.notify_response)



from typing import Any, List

from ..registers import ReadableRegisters, WriteableRegister
from ..fields import DeviceField, EnumField, BoolField, SwitchField, SelectField


class BluettiDevice:
    def __init__(self, fields: List[DeviceField]):
        self.fields = fields

    def addFields(self, fields: List[DeviceField]):
        for f in fields:
            self.fields.append(f)

    def get_polling_registers(self) -> List[ReadableRegisters]:
        result = []

        # TODO Optimize this to reduce amount of registers to separately request

        for f in self.fields:
            result.append(ReadableRegisters(f.address, f.size))

        return result

    def get_full_registers_range(self) -> List[ReadableRegisters]:
        raise NotImplementedError

    def get_device_type_registers(self) -> List[ReadableRegisters]:
        raise NotImplementedError

    def get_device_sn_registers(self) -> List[ReadableRegisters]:
        raise NotImplementedError

    def get_iot_version(self) -> int:
        raise NotImplementedError

    def parse(self, starting_address: int, data: bytes) -> dict:
        # Offsets and size are counted in 2 byte chunks, so for the range we
        # need to divide the byte size by 2
        data_size = int(len(data) / 2)

        # Filter out fields not in range
        r = range(starting_address, starting_address + data_size)
        fields = [
            f for f in self.fields if f.address in r and f.address + f.size - 1 in r
        ]

        # Parse fields
        parsed = {}
        for f in fields:
            data_start = 2 * (f.address - starting_address)
            field_data = data[data_start : data_start + 2 * f.size]
            value = f.parse(field_data)
            if not f.in_range(value):
                continue
            parsed[f.name] = value

        return parsed

    def build_write_command(self, name: str, value: Any) -> WriteableRegister:
        matches = [f for f in self.fields if f.name == name]
        field = next(f for f in matches if f.is_writeable)

        # Convert value to an integer
        if isinstance(field, SelectField):
            value = field.e[value].value
        elif isinstance(field, BoolField):
            value = 1 if value else 0

        return WriteableRegister(field.address, value)

    def get_bool_fields(self):
        return [f for f in self.fields if isinstance(f, BoolField)]

    def get_switch_fields(self):
        return [f for f in self.fields if isinstance(f, SwitchField)]

    def get_select_fields(self):
        return [f for f in self.fields if isinstance(f, SelectField)]

    def get_sensor_fields(self):
        return [
            f
            for f in self.fields
            if not isinstance(f, BoolField)
            and not isinstance(f, SwitchField)
            and not isinstance(f, SelectField)
        ]



from typing import Dict
from .FieldName import FieldName

FIELD_UNIT: Dict[FieldName, str] = {
    FieldName.AC_INPUT_CURRENT: "A",
    FieldName.AC_INPUT_FREQUENCY: "Hz",
    FieldName.AC_INPUT_POWER: "W",
    FieldName.AC_INPUT_VOLTAGE: "V",
    FieldName.AC_OUTPUT_CURRENT: "A",
    FieldName.AC_OUTPUT_FREQUENCY: "Hz",
    FieldName.AC_OUTPUT_POWER: "W",
    FieldName.AC_OUTPUT_VOLTAGE: "V",
    FieldName.AC_P1_CURRENT: "A",
    FieldName.AC_P1_POWER: "W",
    FieldName.AC_P1_VOLTAGE: "V",
    FieldName.AC_P2_CURRENT: "A",
    FieldName.AC_P2_POWER: "W",
    FieldName.AC_P2_VOLTAGE: "V",
    FieldName.AC_P3_CURRENT: "A",
    FieldName.AC_P3_POWER: "W",
    FieldName.AC_P3_VOLTAGE: "V",
    FieldName.BATTERY_SOC: "%",
    FieldName.BATTERY_SOC_RANGE_END: "%",
    FieldName.BATTERY_SOC_RANGE_START: "%",
    FieldName.CTRL_ECO_MIN_POWER_AC: "W",
    FieldName.CTRL_ECO_MIN_POWER_DC: "W",
    FieldName.DC_INPUT_CURRENT: "A",
    FieldName.DC_INPUT_POWER: "W",
    FieldName.DC_INPUT_VOLTAGE: "V",
    FieldName.DC_OUTPUT_POWER: "W",
    FieldName.GRID_FREQUENCY: "Hz",
    FieldName.GRID_P1_CURRENT: "A",
    FieldName.GRID_P1_POWER: "W",
    FieldName.GRID_P1_VOLTAGE: "V",
    FieldName.GRID_P2_CURRENT: "A",
    FieldName.GRID_P2_POWER: "W",
    FieldName.GRID_P2_VOLTAGE: "V",
    FieldName.GRID_P3_CURRENT: "A",
    FieldName.GRID_P3_POWER: "W",
    FieldName.GRID_P3_VOLTAGE: "V",
    FieldName.INTERNAL_AC_FREQUENCY: "Hz",
    FieldName.INTERNAL_AC_VOLTAGE: "V",
    FieldName.INTERNAL_DC_INPUT_CURRENT: "A",
    FieldName.INTERNAL_DC_INPUT_POWER: "W",
    FieldName.INTERNAL_DC_INPUT_VOLTAGE: "V",
    FieldName.POWER_GENERATION: "kWh",
    FieldName.PV_S1_CURRENT: "A",
    FieldName.PV_S1_POWER: "W",
    FieldName.PV_S1_VOLTAGE: "V",
    FieldName.PV_S2_CURRENT: "A",
    FieldName.PV_S2_POWER: "W",
    FieldName.PV_S2_VOLTAGE: "V",
    FieldName.SM_P1_CURRENT: "A",
    FieldName.SM_P1_POWER: "W",
    FieldName.SM_P1_VOLTAGE: "V",
    FieldName.SM_P2_CURRENT: "A",
    FieldName.SM_P2_POWER: "W",
    FieldName.SM_P2_VOLTAGE: "V",
    FieldName.SM_P3_CURRENT: "A",
    FieldName.SM_P3_POWER: "W",
    FieldName.SM_P3_VOLTAGE: "V",
    FieldName.TIME_REMAINING: "h",
}


def get_unit(field: FieldName) -> str | None:
    return FIELD_UNIT.get(field)



from ..base_devices import BaseDeviceV2
from ..fields import FieldName, UIntField, DecimalField, BoolField


class AC2A(BaseDeviceV2):
    def __init__(self):
        super().__init__(
            [
                UIntField(FieldName.DC_OUTPUT_POWER, 140),
                UIntField(FieldName.AC_OUTPUT_POWER, 142),
                UIntField(FieldName.DC_INPUT_POWER, 144),
                UIntField(FieldName.AC_INPUT_POWER, 146),
                DecimalField(FieldName.POWER_GENERATION, 154),
                BoolField(FieldName.AC_OUTPUT_ON, 1509),
                BoolField(FieldName.DC_OUTPUT_ON, 2012),
                BoolField(FieldName.POWER_LIFTING_ON, 2021),
            ],
        )



import json
import argparse
import asyncio
import logging

from ..base_devices import BaseDeviceV1, BaseDeviceV2
from ..fields import FieldName, get_unit
from ..utils.device_builder import build_device
from .types import ReadallData


async def async_parse_file(filename: str):
    print(f"Reading file {filename}")

    with open(filename, "r") as json_data:
        dict_data = json.load(json_data)
        data = ReadallData(**dict_data)

    if data.iotVersion == 1:
        device = BaseDeviceV1()
    else:
        device = BaseDeviceV2()

    registers: list[bytes] = [bytes.fromhex(b) for b in data.registers.values()]

    parsed = {}

    addr = 1
    for r in registers:
        parsed.update(device.parse(addr, r))
        addr += 50

    device_type = parsed.get(FieldName.DEVICE_TYPE.value)

    device = build_device(device_type + "12345678")

    data = {}

    addr = 1
    for r in registers:
        data.update(device.parse(addr, r))
        addr += 50

    print()
    for key, value in data.items():
        key = FieldName(key)
        unit = get_unit(key)
        print(f"{key}: {value}" + ("" if unit is None else unit))


def start():
    """Entrypoint."""
    parser = argparse.ArgumentParser(description="Parse readall output files")
    parser.add_argument(
        "file", type=str, help="JSON file of the powerstation readall output"
    )
    args = parser.parse_args()

    if args.file is None:
        parser.print_help()
        return

    logging.basicConfig(level=logging.DEBUG)

    asyncio.run(async_parse_file(args.file))



import asyncio
import logging
from typing import Any
import async_timeout
from bleak import BleakClient
from bleak.exc import BleakError

from ..fields import FieldName
from ..const import WRITE_UUID
from ..base_devices import BluettiDevice

_LOGGER = logging.getLogger(__name__)


class DeviceWriterConfig:
    def __init__(self, timeout: int = 15, use_encryption: bool = False):
        self.timeout = timeout
        self.use_encryption = use_encryption


class DeviceWriter:
    def __init__(
        self,
        bleak_client: BleakClient,
        bluetti_device: BluettiDevice,
        config: DeviceWriterConfig = DeviceWriterConfig(),
        lock: asyncio.Lock = asyncio.Lock(),
    ):
        self.client = bleak_client
        self.bluetti_device = bluetti_device
        self.config = config
        self.polling_lock = lock

    async def write(self, field: str, value: Any):
        if self.config.use_encryption:
            _LOGGER.error("Encryption on writes is not yet supported")
            return

        available_fields = [f.name for f in self.bluetti_device.fields]
        if field not in available_fields:
            _LOGGER.error("Field not supported")
            return

        command = self.bluetti_device.build_write_command(field, value)

        _LOGGER.debug("Writing to device register")

        async with self.polling_lock:
            try:
                async with async_timeout.timeout(self.config.timeout):
                    if not self.client.is_connected:
                        _LOGGER.debug("Connecting to device")
                        await self.client.connect()

                    _LOGGER.debug("Connected to device")

                    await self.client.write_gatt_char(
                        WRITE_UUID,
                        bytes(command),
                    )

                    # Wait until device has changed value, otherwise reading register might reset it
                    await asyncio.sleep(5)

            except TimeoutError:
                _LOGGER.warning("Timeout")
                return None
            except BleakError as err:
                _LOGGER.warning("Bleak error: %s", err)
                return None
            except BaseException as err:
                _LOGGER.warning("Unknown error: %s", err)
                return None
            finally:
                await self.client.disconnect()
                _LOGGER.debug("Disconnected from device")



from . import DeviceField, FieldName


def swap_bytes(data: bytes):
    """Swaps the place of every other byte, returning a new byte array"""
    arr = bytearray(data)
    for i in range(0, len(arr) - 1, 2):
        arr[i], arr[i + 1] = arr[i + 1], arr[i]
    return arr


class SwapStringField(DeviceField):
    def __init__(self, name: FieldName, address: int, size: int):
        super().__init__(name, address, size)

    def parse(self, data: bytes) -> str:
        return swap_bytes(data).rstrip(b"\0").decode("ascii", errors="ignore")



import struct

from . import DeviceRegister, RegisterAction


class WriteableRegister(DeviceRegister):
    def __init__(self, address: int, value: int):
        super().__init__(RegisterAction.WRITE, struct.pack("!HH", address, value))
        self.address = address
        self.value = value

    def response_size(self):
        return 8

    def parse_response(self, response: bytes):
        return bytes(response[4:6])

    def __repr__(self):
        return f"WriteableRegister(address={self.address}, value={self.value:#04x})"



import struct
from enum import Enum
from typing import Any, Type

from . import DeviceField, FieldName


class EnumField(DeviceField):
    def __init__(self, name: FieldName, address: int, enum: Type[Enum]):
        super().__init__(name, address, 1)
        self.enum = enum

    def parse(self, data: bytes) -> Any:
        val = struct.unpack("!H", data)[0]
        return self.enum(val)



"""Bluetti scan command."""

import argparse
import asyncio
import logging
from typing import List
from bleak import BleakScanner
from bleak.backends.device import BLEDevice

from ..utils.device_info import get_type_by_bt_name


async def scan_async():
    stop_event = asyncio.Event()

    found: List[List[str]] = []

    async def callback(device: BLEDevice, _):
        result = get_type_by_bt_name(device.name)

        if result is None:
            return

        if result is not None or device.name.startswith("PBOX"):
            found.append(device.address)
            stop_event.set()
            print([result, device.address])

    async with BleakScanner(callback):
        await stop_event.wait()


def start():
    """Entrypoint."""
    parser = argparse.ArgumentParser(
        description="Detect bluetti devices by bluetooth name"
    )
    parser.parse_args()

    # logging.basicConfig(level=logging.DEBUG)

    asyncio.run(scan_async())



from ..base_devices import BaseDeviceV2
from ..enums import ChargingMode, EcoMode
from ..fields import FieldName, UIntField, DecimalField, SwitchField, SelectField


class AC180(BaseDeviceV2):
    def __init__(self):
        super().__init__(
            [
                UIntField(FieldName.DC_OUTPUT_POWER, 140),
                UIntField(FieldName.AC_OUTPUT_POWER, 142),
                UIntField(FieldName.DC_INPUT_POWER, 144),
                UIntField(FieldName.AC_INPUT_POWER, 146),
                DecimalField(FieldName.DC_INPUT_VOLTAGE, 1213, 1),
                DecimalField(FieldName.DC_INPUT_CURRENT, 1214, 1),
                DecimalField(FieldName.AC_INPUT_FREQUENCY, 1300, 1),
                UIntField(FieldName.AC_INPUT_VOLTAGE, 1314, 0.1),
                DecimalField(FieldName.AC_INPUT_CURRENT, 1315, 1),
                DecimalField(FieldName.AC_OUTPUT_FREQUENCY, 1500, 1),
                DecimalField(FieldName.AC_OUTPUT_VOLTAGE, 1511, 1),
                SwitchField(FieldName.CTRL_AC, 2011),
                SwitchField(FieldName.CTRL_DC, 2012),
                SwitchField(FieldName.CTRL_ECO_DC, 2014),
                SelectField(FieldName.CTRL_ECO_TIME_MODE_DC, 2015, EcoMode),
                UIntField(FieldName.CTRL_ECO_MIN_POWER_DC, 2016),  # Not controlable
                SwitchField(FieldName.CTRL_ECO_AC, 2017),
                SelectField(FieldName.CTRL_ECO_TIME_MODE_AC, 2018, EcoMode),
                UIntField(FieldName.CTRL_ECO_MIN_POWER_AC, 2019),  # Not controlable
                SelectField(FieldName.CTRL_CHARGING_MODE, 2020, ChargingMode),
                SwitchField(FieldName.CTRL_POWER_LIFTING, 2021),
                # Display timeout (?)
                # AC frequency (?)
                # Grid stabilizing (?)
            ],
        )



import json

from dataclasses import dataclass


@dataclass
class ReadallData:
    mac: str
    iotVersion: int
    encryption: bool
    registers: dict[str, str]

    def toJSON(self):
        return {
            "mac": self.mac,
            "iotVersion": self.iotVersion,
            "encryption": self.encryption,
            "registers": self.registers,
        }



from enum import Enum, unique


@unique
class FieldName(Enum):
    AC_INPUT_CURRENT = "ac_input_current"
    AC_INPUT_FREQUENCY = "ac_input_frequency"
    AC_INPUT_POWER = "ac_input_power"
    AC_INPUT_VOLTAGE = "ac_input_voltage"
    AC_OUTPUT_FREQUENCY = "ac_output_frequency"
    AC_OUTPUT_MODE = "ac_output_mode"
    AC_OUTPUT_ON = "ac_output_on"
    AC_OUTPUT_CURRENT = "ac_output_current"
    AC_OUTPUT_POWER = "ac_output_power"
    AC_OUTPUT_VOLTAGE = "ac_output_voltage"
    AC_P1_CURRENT = "ac_p1_current"
    AC_P1_POWER = "ac_p1_power"
    AC_P1_VOLTAGE = "ac_p1_voltage"
    AC_P2_CURRENT = "ac_p2_current"
    AC_P2_POWER = "ac_p2_power"
    AC_P2_VOLTAGE = "ac_p2_voltage"
    AC_P3_CURRENT = "ac_p3_current"
    AC_P3_POWER = "ac_p3_power"
    AC_P3_VOLTAGE = "ac_p3_voltage"
    BATTERY_SOC = "total_battery_percent"
    BATTERY_SOC_RANGE_END = "soc_range_end"
    BATTERY_SOC_RANGE_START = "soc_range_start"
    CTRL_AC = "ctrl_ac"
    CTRL_CHARGING_MODE = "ctrl_charging_mode"
    CTRL_DC = "ctrl_dc"
    CTRL_DISPLAY_TIMEOUT = "ctrl_display_timeout"
    CTRL_ECO = "ctrl_eco"
    CTRL_ECO_AC = "ctrl_eco_ac"
    CTRL_ECO_DC = "ctrl_eco_dc"
    CTRL_ECO_MIN_POWER_AC = "ctrl_eco_min_power_ac"
    CTRL_ECO_MIN_POWER_DC = "ctrl_eco_min_power_dc"
    CTRL_ECO_TIME_MODE = "ctrl_eco_time_mode"
    CTRL_ECO_TIME_MODE_AC = "ctrl_eco_time_mode_ac"
    CTRL_ECO_TIME_MODE_DC = "ctrl_eco_time_mode_dc"
    CTRL_GRID_STABILIZING = "ctrl_grid_stabilizing"
    CTRL_LED_MODE = "ctrl_led_mode"
    CTRL_POWER_LIFTING = "ctrl_power_lifting"
    CTRL_POWER_OFF = "ctrl_power_off"
    DC_INPUT_CURRENT = "dc_input_current"
    DC_INPUT_POWER = "dc_input_power"
    DC_INPUT_VOLTAGE = "dc_input_voltage"
    DC_OUTPUT_ON = "dc_output_on"
    DC_OUTPUT_POWER = "dc_output_power"
    DEVICE_TYPE = "device_type"
    DEVICE_SN = "device_sn"
    GRID_FREQUENCY = "grid_frequency"
    GRID_P1_CURRENT = "grid_p1_current"
    GRID_P1_POWER = "grid_p1_power"
    GRID_P1_VOLTAGE = "grid_p1_voltage"
    GRID_P2_CURRENT = "grid_p2_current"
    GRID_P2_POWER = "grid_p2_power"
    GRID_P2_VOLTAGE = "grid_p2_voltage"
    GRID_P3_CURRENT = "grid_p3_current"
    GRID_P3_POWER = "grid_p3_power"
    GRID_P3_VOLTAGE = "grid_p3_voltage"
    INTERNAL_AC_FREQUENCY = "internal_ac_frequency"
    INTERNAL_AC_VOLTAGE = "internal_ac_voltage"
    INTERNAL_DC_INPUT_CURRENT = "internal_dc_input_current"
    INTERNAL_DC_INPUT_POWER = "internal_dc_input_power"
    INTERNAL_DC_INPUT_VOLTAGE = "internal_dc_input_voltage"
    POWER_GENERATION = "power_generation"
    POWER_LIFTING_ON = "power_lifting_on"
    PV_S1_CURRENT = "pv_s1_current"
    PV_S1_POWER = "pv_s1_power"
    PV_S1_VOLTAGE = "pv_s1_voltage"
    PV_S2_CURRENT = "pv_s2_current"
    PV_S2_POWER = "pv_s2_power"
    PV_S2_VOLTAGE = "pv_s2_voltage"
    SM_P1_CURRENT = "sm_p1_current"  # Smart meter
    SM_P1_POWER = "sm_p1_power"
    SM_P1_VOLTAGE = "sm_p1_voltage"
    SM_P2_CURRENT = "sm_p2_current"
    SM_P2_POWER = "sm_p2_power"
    SM_P2_VOLTAGE = "sm_p2_voltage"
    SM_P3_CURRENT = "sm_p3_current"
    SM_P3_POWER = "sm_p3_power"
    SM_P3_VOLTAGE = "sm_p3_voltage"
    TIME_REMAINING = "time_remaining"
    VER_ARM = "version_arm"
    VER_DSP = "version_dsp"



import argparse
import asyncio
import json
import logging

from ..bluetooth.device_reader import DeviceReader, DeviceReaderConfig
from ..base_devices import BaseDeviceV1, BaseDeviceV2
from .types import ReadallData


async def async_read_device(address: str, iot_version: int, encryption: bool):
    device = None
    match iot_version:
        case 1:
            device = BaseDeviceV1()
        case 2:
            device = BaseDeviceV2()

    if device is None:
        print("Unknown iot protocol version")
        return

    reader = DeviceReader(
        address, device, asyncio.Future, DeviceReaderConfig(use_encryption=encryption)
    )

    print("Reader created")

    data = await reader.read(
        only_registers=device.get_full_registers_range(),
        raw=True,
    )

    if data is None:
        print("Error")
        return

    print("Writing data to file")

    register_data = {}
    for key, value in data.items():
        register_data[key] = value.hex()

    data_obj = ReadallData(
        address,
        iot_version,
        encryption,
        register_data,
    )

    with open(f"bluetti_data.{address.replace(':', '-')}.json", "w") as f:
        json.dump(data_obj.toJSON(), f)


def start():
    """Entrypoint."""
    parser = argparse.ArgumentParser(description="Detect bluetti devices")
    parser.add_argument("-m", "--mac", type=str, help="Mac-address of the powerstation")
    parser.add_argument("-v", "--version", type=int, help="IoT protocol version")
    parser.add_argument(
        "-e", "--encryption", type=bool, help="Add this if encryption is needed"
    )
    args = parser.parse_args()

    if args.mac is None or args.version is None:
        parser.print_help()
        return

    encryption = False if args.encryption is None else True

    # logging.basicConfig(level=logging.DEBUG)

    asyncio.run(async_read_device(args.mac, args.version, encryption))



from typing import List

from . import BluettiDevice
from ..fields import DeviceField
from ..fields import FieldName, SwapStringField, UIntField, SerialNumberField
from ..registers import ReadableRegisters


class BaseDeviceV2(BluettiDevice):
    def __init__(self, additional_fields: List[DeviceField] = []):
        super().__init__(
            [
                SwapStringField(FieldName.DEVICE_TYPE, 110, 6),
                SerialNumberField(FieldName.DEVICE_SN, 116),
                UIntField(FieldName.BATTERY_SOC, 102, min=0, max=100),
            ]
            + additional_fields,
        )

    def get_full_registers_range(self) -> List[ReadableRegisters]:
        return [ReadableRegisters(i, 50) for i in range(1, 8000, 50)]

    def get_device_type_registers(self) -> List[ReadableRegisters]:
        return [
            ReadableRegisters(110, 6),
        ]

    def get_device_sn_registers(self) -> List[ReadableRegisters]:
        return [
            ReadableRegisters(116, 4),
        ]

    def get_iot_version(self) -> int:
        return 2



import re

from .ac2a import AC2A
from .ac50b import AC50B
from .ac60 import AC60
from .ac60p import AC60P
from .ac70 import AC70
from .ac70p import AC70P
from .ac180 import AC180
from .ac180p import AC180P
from .ac180t import AC180T
from .ac200l import AC200L
from .ac200m import AC200M
from .ac200pl import AC200PL
from .ac300 import AC300
from .ac500 import AC500
from .ap300 import AP300
from .eb3a import EB3A
from .el100v2 import EL100V2
from .el30v2 import EL30V2
from .ep500 import EP500
from .ep500p import EP500P
from .ep600 import EP600
from .ep800 import EP800
from .handsfree1 import Handsfree1

# Add new device classes here
DEVICES = {
    "AC2A": AC2A,
    "AC50B": AC50B,
    "AC60": AC60,
    "AC60P": AC60P,
    "AC70": AC70,
    "AC70P": AC70P,
    "AC180": AC180,
    "AC180T": AC180T,
    "AC180P": AC180P,
    "AC200L": AC200L,
    "AC200M": AC200M,
    "AC200PL": AC200PL,
    "AC300": AC300,
    "AC500": AC500,
    "AP300": AP300,
    "EB3A": EB3A,
    "EL100V2": EL100V2,
    "EL30V2": EL30V2,
    "EP500": EP500,
    "EP500P": EP500P,
    "EP600": EP600,
    "EP800": EP800,
    "Handsfree 1": Handsfree1,
}

# Prefixes of all currently supported devices
DEVICE_NAME_RE = re.compile(
    r"^(AC2A|AC50B|AC60|AC60P|AC70|AC70P|AC180|AC180T|AC180P|AC200L|AC200M|AC200PL|AC300|AC500|AP300|EB3A|EL100V2|EL30V2|EP500|EP500P|EP600|EP760|EP800|Handsfree\s1)(\d+)$"
)



import struct
from decimal import Decimal

from . import DeviceField, FieldName


class DecimalField(DeviceField):
    def __init__(
        self,
        name: FieldName,
        address: int,
        scale: int = 1,
        multiplier: float = 1,
        min: Decimal | None = None,
        max: Decimal | None = None,
    ):
        super().__init__(name, address, 1)
        self.scale = scale
        self.multiplier = multiplier
        self.min = min
        self.max = max

    def parse(self, data: bytes) -> Decimal:
        val = Decimal(struct.unpack("!H", data)[0])
        return (val / 10 ** self.scale) * Decimal(self.multiplier)

    def in_range(self, value: Decimal) -> bool:
        if self.min is not None and self.min > value:
            return False
        if self.max is not None and self.max < value:
            return False
        return True



from ..base_devices import BaseDeviceV2
from ..fields import FieldName, UIntField, DecimalField


class AC70P(BaseDeviceV2):
    def __init__(self):
        super().__init__(
            [
                UIntField(FieldName.DC_OUTPUT_POWER, 140),
                UIntField(FieldName.AC_OUTPUT_POWER, 142),
                UIntField(FieldName.DC_INPUT_POWER, 144),
                UIntField(FieldName.AC_INPUT_POWER, 146),
                DecimalField(FieldName.DC_INPUT_VOLTAGE, 1213, 1),
                DecimalField(FieldName.DC_INPUT_CURRENT, 1214, 1),
                DecimalField(FieldName.AC_INPUT_FREQUENCY, 1300, 1),
                UIntField(FieldName.AC_INPUT_VOLTAGE, 1314, 0.1),
                DecimalField(FieldName.AC_INPUT_CURRENT, 1315, 1),
                DecimalField(FieldName.AC_OUTPUT_FREQUENCY, 1500, 1),
                DecimalField(FieldName.AC_OUTPUT_VOLTAGE, 1511, 1),
            ],
        )



import struct

from . import DeviceRegister, RegisterAction


class ReadableRegisters(DeviceRegister):
    def __init__(self, starting_address: int, quantity: int):
        super().__init__(
            RegisterAction.READ, struct.pack("!HH", starting_address, quantity)
        )
        self.starting_address = starting_address
        self.quantity = quantity

    def response_size(self):
        # 3 byte header
        # each returned field is actually 2 bytes (16-bit word)
        # 2 byte crc
        return 2 * self.quantity + 5

    def parse_response(self, response: bytes):
        return bytes(response[3:-2])

    def __repr__(self):
        return f"ReadableRegisters(starting_address={self.starting_address}, quantity={self.quantity})"



"""Device info helper."""

from ..devices import DEVICE_NAME_RE


def get_type_by_bt_name(bt_name: str) -> str | None:
    """Check bluetooth name and return type if supported."""

    # Some devices don't show up with a name.  re.match() will fail on None type
    if bt_name is None:
        return None

    match = DEVICE_NAME_RE.match(bt_name)

    if match is None:
        return None

    return match[1]



from enum import Enum
import hashlib
import logging
import os
import pyasn1.codec.der.decoder as der_decoder
import pyasn1.codec.der.encoder as der_encoder
from pyasn1.type import univ
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.exceptions import InvalidSignature

LOCAL_AES_KEY = "459FC535808941F17091E0993EE3E93D"
PRIVATE_KEY_L1 = "4F19A16E3E87BDD9BD24D3E5495B88041511943CBC8B969ADE9641D0F56AF337"
PUBLIC_KEY_K2 = "3059301306072a8648ce3d020106082a8648ce3d03010703420004A73ABF5D2232C8C1C72E68304343C272495E3A8FD6F30EA96DE2F4B3CE60B251EE21AC667CF8A71E18B46B664EAEFFE3C489F24F695B6411DB7E22CCC85A8594"
SECP_256R1_PUBLIC_PREFIX = "3059301306072a8648ce3d020106082a8648ce3d03010703420004"

_LOGGER = logging.getLogger(__name__)

KEX_MAGIC = b"**"
AES_BLOCK_SIZE = 16


def hexsum(s, sz: int):
    checksum = sum(s)
    as_hex = f"{checksum:0{sz*2}x}"
    return bytes.fromhex(as_hex)


def hexxor(a: bytes, b: bytes) -> bytes | None:
    if len(a) != len(b):
        _LOGGER.error("Can only XOR two identical length byte strings")
        return None
    return bytes([x ^ y for x, y in zip(a, b)])


def raw_ecdsa_to_der(sig):
    # <byte r[32]> <byte s[32]>

    if len(sig) != 64:
        raise ValueError("ecdsa signature is the wrong size")

    seq = univ.SequenceOf()
    seq.extend(
        [
            univ.Integer(int.from_bytes(sig[:32], "big")),  # r
            univ.Integer(int.from_bytes(sig[32:], "big")),  # s
        ]
    )
    return der_encoder.encode(seq)


def der_to_raw_ecdsa(sig):
    # 30 45 02 20 1956307e59448178b47c222e4e1e6c8ef7d707bc230e5a9fa77f919ec44e5f74
    # |  |  |  |  |> byte r[0x20]
    # |  |  |  |---> Length
    # |  |  |------> DER type (int)
    # |  |---------> Payload size
    # |------------> DER type (sequence)

    #       02 21 00cfad6e11abd5e803fb6874c3838bc968db1f3c070ae6b85db9d8ed8936a1eb5c
    #       |  |  |> byte s[0x21]
    #       |  |---> Length
    #       |------> DER type (int)

    seq, remainder = der_decoder.decode(sig)
    if remainder:
        raise ValueError("Found trailing data")

    return b"".join([int.to_bytes(int(x), 0x20, "big") for x in seq])


def verify_and_extract_signed_data(message, signed_data_suffix: bytes | None):
    # 64 bytes of data
    # 64 bytes of signature
    if len(message) != 128:
        raise ValueError("Unexpected message length")

    data = message[:64]
    signature = message[64:]
    signed_data = data.tobytes() + signed_data_suffix
    der_signature = raw_ecdsa_to_der(signature)
    try:
        key_bytes = bytes.fromhex(PUBLIC_KEY_K2)
        serialization.load_der_public_key(key_bytes).verify(
            der_signature, signed_data, ec.ECDSA(hashes.SHA256())
        )
        _LOGGER.debug("Signature OK")
    except InvalidSignature:
        raise

    return data


def pubkey_from_bytes(data):
    encoded_peer_pubkey = bytes.fromhex(SECP_256R1_PUBLIC_PREFIX) + data
    return serialization.load_der_public_key(encoded_peer_pubkey)


def generate_keypair():
    private = ec.generate_private_key(ec.SECP256R1())
    return (private.public_key(), private)


def pubkey_to_bytes(pubkey):
    out = pubkey.public_bytes(
        encoding=serialization.Encoding.X962,
        format=serialization.PublicFormat.UncompressedPoint,
    )

    if out[0] != 0x4 or len(out) != 65:
        raise RuntimeError(
            "First byte should be 04 for the uncompressed format, and total size 64"
        )

    return out[1:]


class MessageType(Enum):
    CHALLENGE = 1
    CHALLENGE_ACCEPTED = 3
    PEER_PUBKEY = 4
    PUBKEY_ACCEPTED = 6


class Message:
    """
    Two types of messages. The first one is for communication that happens before we have
    a proper symmetric key available. These are fully handled by this file, it's purely
    to setup the encryption protocol.
    2a2a .... 0000
    |         |> Body checksum
    |    |-----> Body
    |----------> Magic value for pre-key-exchange
    The second type is for anything after. This can either contain the type 1 message above, or
    anything else -- at this point we're just wrapping the regular protocol, and we just forward
    the packets dowm the regular processing path after decrypting them.
    See aes_decrypt() for the format.
    """

    def __init__(self, buffer: bytes):
        self.buffer = buffer
        self.view = memoryview(self.buffer)

    @property
    def header(self) -> memoryview:
        return self.view[:2]

    @property
    def is_pre_key_exchange(self) -> bool:
        return self.header == KEX_MAGIC

    @property
    def checksum(self) -> memoryview:
        return self.view[-2:]

    @property
    def body(self) -> memoryview:
        return self.view[len(self.header) : -len(self.checksum)]

    @property
    def data(self) -> memoryview:
        return self.body[2:]

    @property
    def type(self) -> int:
        return MessageType(self.body[0])

    def verify_checksum(self):
        message_checksum = self.checksum
        computed_checksum = hexsum(self.body, len(message_checksum))
        if computed_checksum != message_checksum:
            _LOGGER.error("Checksum error!")
        _LOGGER.debug("Checksum OK")


class BluettiEncryption:
    # Derived exclusively from data sent over the network
    # Used for the initial handshake
    unsecure_aes_key: bytes | None = None

    # Predictably derived from a seed sent by the peer
    # This is the same for all the messages encrypted
    # with that key during the connection
    unsecure_aes_iv: bytes | None = None

    # Proper key exchange gives us another key,
    # that is used for the remainder of the connection
    # IV is random per message
    secure_aes_key: bytes | None = None

    # Received through key exchange
    # The signing key for the key exchange is well-known
    peer_pubkey: bytes | None = None

    @property
    def is_ready_for_commands(self) -> bool:
        return self.secure_aes_key is not None and self.peer_pubkey is not None

    def aes_decrypt(self, data: bytes, aes_key: bytes | None, iv: bytes | None):
        data_len = (data[0] << 8) + data[1]

        if iv is None:
            iv = hashlib.md5(data[2:6]).digest()
            encrypted = memoryview(data)[6:]
        else:
            encrypted = memoryview(data[2:])

        if len(encrypted) % AES_BLOCK_SIZE != 0:
            raise ValueError("Data not aligned on aes block size")

        cipher = Cipher(algorithms.AES(aes_key), modes.CBC(iv))
        decryptor = cipher.decryptor()
        decrypted = decryptor.update(encrypted) + decryptor.finalize()
        decrypted = decrypted[:data_len]

        _LOGGER.debug(">PLAIN " + decrypted.hex())
        return decrypted

    def aes_encrypt(self, data: bytes, aes_key: bytes | None, iv: bytes | None):
        message_header = int.to_bytes(len(data), 2, "big")

        if iv is None:
            iv_seed = os.urandom(4)
            iv = hashlib.md5(iv_seed).digest()
            message_header += iv_seed

        padding = (AES_BLOCK_SIZE - len(data) % AES_BLOCK_SIZE) % AES_BLOCK_SIZE
        data += bytes(padding)

        cipher = Cipher(algorithms.AES(aes_key), modes.CBC(iv))
        encryptor = cipher.encryptor()
        encrypted = encryptor.update(data) + encryptor.finalize()
        encrypted = message_header + encrypted

        _LOGGER.debug("PLAIN> " + data.hex())
        return encrypted

    def msg_challenge(self, message: Message) -> bytes | None:
        _LOGGER.debug("Received challenge")

        if len(message.data) != 4:
            _LOGGER.error("Unexpected message length")
            return None

        self.unsecure_aes_iv = hashlib.md5(message.data[::-1].tobytes()).digest()
        static_key = bytes.fromhex(LOCAL_AES_KEY)
        self.unsecure_aes_key = hexxor(self.unsecure_aes_iv, static_key)

        _LOGGER.info("Unsecure iv  " + self.unsecure_aes_iv.hex())
        _LOGGER.info("Unsecure key " + self.unsecure_aes_key.hex())

        body = bytes.fromhex("0204") + self.unsecure_aes_iv[8:12]
        return b"".join([KEX_MAGIC, body, hexsum(body, 2)])

    def msg_peer_pubkey(self, message: Message) -> bytes | None:
        _LOGGER.debug("Received peer pubkey, checking signature")
        data = verify_and_extract_signed_data(message.data, self.unsecure_aes_iv)
        self.peer_pubkey = pubkey_from_bytes(data)

        _LOGGER.debug("Generating a local keypair")
        self.my_pubkey, self.my_privkey = generate_keypair()
        my_pubkey_bytes = pubkey_to_bytes(self.my_pubkey)

        _LOGGER.debug("Signing the local pubkey")
        signing_secret = int.from_bytes(bytes.fromhex(PRIVATE_KEY_L1), "big")
        signing_key = ec.derive_private_key(signing_secret, ec.SECP256R1())
        to_sign = my_pubkey_bytes + self.unsecure_aes_iv
        signature = signing_key.sign(to_sign, ec.ECDSA(hashes.SHA256()))
        raw_signature = der_to_raw_ecdsa(signature)

        body = b"".join([bytes.fromhex("0580"), my_pubkey_bytes, raw_signature])
        msg = b"".join([KEX_MAGIC, body, hexsum(body, 2)])
        return self.aes_encrypt(msg, self.unsecure_aes_key, self.unsecure_aes_iv)

    def msg_key_accepted(self, message: Message) -> None:
        _LOGGER.debug("Received key exchange confirmation, calculating shared secret")

        if len(message.data) != 1:
            raise ValueError("Unexpected message length")
        if message.data[0] != 0:
            raise ValueError("Key acceptance response is not 0")

        self.secure_aes_key = self.my_privkey.exchange(ec.ECDH(), self.peer_pubkey)
        _LOGGER.info("Secure key   " + self.secure_aes_key.hex())

    def getKeyIv(self):
        return (
            (self.unsecure_aes_key, self.unsecure_aes_iv)
            if self.secure_aes_key is None
            else (self.secure_aes_key, None)
        )

    def reset(self):
        self.peer_pubkey = None
        self.secure_aes_key = None



from typing import Any

from ..fields import FieldName


class DeviceField:
    def __init__(self, name: FieldName, address: int, size: int):
        self.name = name.value
        self.address = address
        self.size = size

    def parse(self, data: bytes) -> Any:
        raise NotImplementedError

    def is_writeable(self) -> bool:
        return False

    def allowed_write_type(self, value: Any) -> bool:
        return False

    def in_range(self, value: Any) -> bool:
        return True



import struct

from . import DeviceField, FieldName


class SerialNumberField(DeviceField):
    def __init__(self, name: FieldName, address: int):
        super().__init__(name, address, 4)

    def parse(self, data: bytes) -> str:
        values = struct.unpack("!4H", data)
        return values[0] + (values[1] << 16) + (values[2] << 32) + (values[3] << 48)



import struct
from decimal import Decimal

from . import DeviceField, FieldName


class VersionField(DeviceField):
    def __init__(self, name: FieldName, address: int):
        super().__init__(name, address, 2)

    def parse(self, data: bytes) -> int:
        values = struct.unpack("!2H", data)
        return Decimal(values[0] + (values[1] << 16)) / 100



from ..base_devices import BaseDeviceV1
from ..fields import FieldName, SwitchField, SelectField, VersionField
from ..enums import EcoMode, LedMode, ChargingMode


class EB3A(BaseDeviceV1):
    def __init__(self):
        super().__init__(
            [
                VersionField(FieldName.VER_ARM, 23),
                VersionField(FieldName.VER_DSP, 25),
                SwitchField(FieldName.CTRL_AC, 3007),
                SwitchField(FieldName.CTRL_DC, 3008),
                SelectField(FieldName.CTRL_LED_MODE, 3034, LedMode),
                SwitchField(FieldName.CTRL_POWER_OFF, 3060),
                SwitchField(FieldName.CTRL_ECO, 3063),
                SelectField(FieldName.CTRL_ECO_TIME_MODE, 3064, EcoMode),
                SelectField(FieldName.CTRL_CHARGING_MODE, 3065, ChargingMode),
                SwitchField(FieldName.CTRL_POWER_LIFTING, 3066),
            ],
        )



"""Bluetti write command."""

import argparse
import asyncio
import logging
from typing import Any
from bleak import BleakClient

from ..bluetooth import DeviceWriter
from ..utils.device_builder import build_device


async def async_write(
    address: str, type: str, encryption: bool, field: str, value: Any
):
    client = BleakClient(address)
    built = build_device(type + "12345678")

    if built is None:
        print("Unsupported powerstation type")
        return

    if encryption:
        print("Encryption is not supported")
        return

    print("Client created")

    writer = DeviceWriter(
        client,
        built,
    )

    print("Writer created")

    await writer.write(field, value)


def start():
    """Entrypoint."""
    parser = argparse.ArgumentParser(description="Write to bluetti device")
    parser.add_argument("-m", "--mac", type=str, help="Mac-address of the powerstation")
    parser.add_argument(
        "-t", "--type", type=str, help="Type of the powerstation (AC70 f.ex.)"
    )
    parser.add_argument("--on", type=bool, help="Value to write")
    parser.add_argument("--off", type=bool, help="Value to write")
    parser.add_argument("-v", "--value", type=int, help="Value to write")
    parser.add_argument(
        "-e", "--encryption", type=bool, help="Add this if encryption is needed"
    )
    parser.add_argument("field", type=str, help="Field name (ctrl_dc f.ex.)")
    args = parser.parse_args()

    if args.mac is None or args.type is None or args.field is None:
        parser.print_help()
        return

    # No value given
    if args.on is None and args.off is None and args.value is None:
        parser.print_help()
        return

    value = args.on is not None
    if args.off is not None:
        value = False

    if args.value:
        value = args.value

    # logging.basicConfig(level=logging.DEBUG)

    asyncio.run(
        async_write(
            args.mac,
            args.type,
            args.encryption,
            args.field,
            value,
        )
    )



from .DeviceField import *
from .FieldName import *
from .FieldUnit import *

from .BoolField import *
from .DecimalField import *
from .EnumField import *
from .IntField import *
from .SelectField import *
from .SerialNumberField import *
from .StringField import *
from .SwapStringField import *
from .SwitchField import *
from .UIntField import *
from .VersionField import *



import struct
import crcmod.predefined
from enum import Enum

modbus_crc = crcmod.predefined.mkCrcFun("modbus")


class RegisterAction(Enum):
    READ = 3
    WRITE = 6


class DeviceRegister:
    def __init__(self, register_action: RegisterAction, data: bytes):
        self.register_action = register_action

        self.cmd = bytearray(len(data) + 4)
        self.cmd[0] = 1
        self.cmd[1] = register_action.value
        self.cmd[2:-2] = data
        struct.pack_into("<H", self.cmd, -2, modbus_crc(self.cmd[:-2]))

    def response_size(self) -> int:
        """Returns the expected response size in bytes"""
        raise NotImplementedError()

    def __iter__(self):
        """Provide an iter implemention so that bytes(cmd) works"""
        return iter(self.cmd)

    def is_exception_response(self, response: bytes):
        """Checks the response code to see if it's a MODBUS exception"""
        if len(response) < 2:
            return False
        else:
            return response[1] == self.register_action.value + 0x80

    def is_valid_response(self, response: bytes):
        """Validates that the reponse is complete and uncorrupted"""
        if len(response) < 3:
            return False

        crc = modbus_crc(response[0:-2])
        crc_bytes = crc.to_bytes(2, byteorder="little")
        return response[-2:] == crc_bytes

    def parse_response(self, response: bytes):
        """Returns the raw body of the response"""
        raise NotImplementedError()



"""Bluetti read command."""

import argparse
import asyncio
import logging

from ..bluetooth.device_reader import DeviceReader, DeviceReaderConfig
from ..utils.device_builder import build_device
from ..fields import FieldName, get_unit


async def async_read_device(address: str, type: str, encryption: bool):
    built = build_device(type + "12345678")

    if built is None:
        print("Unsupported powerstation type")
        return

    print("Client created")

    reader = DeviceReader(
        address, built, asyncio.Future, DeviceReaderConfig(use_encryption=encryption)
    )

    print("Reader created")

    data = await reader.read()

    if data is None:
        print("Error")
        return

    print()
    for key, value in data.items():
        key = FieldName(key)
        unit = get_unit(key)
        print(f"{key}: {value}" + ("" if unit is None else unit))


def start():
    """Entrypoint."""
    parser = argparse.ArgumentParser(description="Detect bluetti devices")
    parser.add_argument("-m", "--mac", type=str, help="Mac-address of the powerstation")
    parser.add_argument(
        "-t", "--type", type=str, help="Type of the powerstation (AC70 f.ex.)"
    )
    parser.add_argument(
        "-e", "--encryption", type=bool, help="Add this if encryption is needed"
    )
    args = parser.parse_args()

    if args.mac is None or args.type is None:
        parser.print_help()
        return

    # logging.basicConfig(level=logging.DEBUG)

    asyncio.run(async_read_device(args.mac, args.type, args.encryption))



from ..base_devices import BaseDeviceV1
from ..fields import FieldName, EnumField, DecimalField, UIntField
from ..enums import OutputMode


class AC200PL(BaseDeviceV1):
    def __init__(self):
        super().__init__(
            [
                EnumField(FieldName.AC_OUTPUT_MODE, 70, OutputMode),
                DecimalField(FieldName.INTERNAL_AC_VOLTAGE, 71, 1, 10),
                DecimalField(FieldName.INTERNAL_AC_FREQUENCY, 74, 2, 10),
                UIntField(FieldName.INTERNAL_DC_INPUT_VOLTAGE, 86, 0.1),
                DecimalField(FieldName.INTERNAL_DC_INPUT_POWER, 87, 1, 10),
                DecimalField(FieldName.INTERNAL_DC_INPUT_CURRENT, 88, 2),
            ],
        )



from ..base_devices import BaseDeviceV2
from ..fields import FieldName, UIntField, DecimalField, IntField


class EP600(BaseDeviceV2):
    def __init__(self):
        super().__init__(
            [
                DecimalField(FieldName.POWER_GENERATION, 1202, 1),
                UIntField(FieldName.PV_S1_POWER, 1212),
                DecimalField(FieldName.PV_S1_VOLTAGE, 1213, 1),
                DecimalField(FieldName.PV_S1_CURRENT, 1214, 1),
                UIntField(FieldName.PV_S2_POWER, 1220),
                DecimalField(FieldName.PV_S2_VOLTAGE, 1221, 1),
                DecimalField(FieldName.PV_S2_CURRENT, 1222, 1),
                UIntField(FieldName.SM_P1_POWER, 1228),
                DecimalField(FieldName.SM_P1_VOLTAGE, 1229, 1),
                UIntField(FieldName.SM_P1_CURRENT, 1230, 1),
                UIntField(FieldName.SM_P2_POWER, 1236),
                DecimalField(FieldName.SM_P2_VOLTAGE, 1237, 1),
                UIntField(FieldName.SM_P2_CURRENT, 1238, 1),
                UIntField(FieldName.SM_P3_POWER, 1244),
                DecimalField(FieldName.SM_P3_VOLTAGE, 1245, 1),
                UIntField(FieldName.SM_P3_CURRENT, 1246, 1),
                DecimalField(FieldName.GRID_FREQUENCY, 1300, 1),
                UIntField(FieldName.GRID_P1_POWER, 1313),
                DecimalField(FieldName.GRID_P1_VOLTAGE, 1314, 1),
                DecimalField(FieldName.GRID_P1_CURRENT, 1315, 1),
                UIntField(FieldName.GRID_P2_POWER, 1319),
                DecimalField(FieldName.GRID_P2_VOLTAGE, 1320, 1),
                DecimalField(FieldName.GRID_P2_CURRENT, 1321, 1),
                UIntField(FieldName.GRID_P3_POWER, 1325),
                DecimalField(FieldName.GRID_P3_VOLTAGE, 1326, 1),
                DecimalField(FieldName.GRID_P3_CURRENT, 1327, 1),
                DecimalField(FieldName.AC_OUTPUT_FREQUENCY, 1500, 1),
                IntField(FieldName.AC_P1_POWER, 1510),
                DecimalField(FieldName.AC_P1_VOLTAGE, 1511, 1),
                DecimalField(FieldName.AC_P1_CURRENT, 1512, 1),
                IntField(FieldName.AC_P2_POWER, 1517),
                DecimalField(FieldName.AC_P2_VOLTAGE, 1518, 1),
                DecimalField(FieldName.AC_P2_CURRENT, 1519, 1),
                IntField(FieldName.AC_P3_POWER, 1524),
                DecimalField(FieldName.AC_P3_VOLTAGE, 1525, 1),
                DecimalField(FieldName.AC_P3_CURRENT, 1526, 1),
                IntField(FieldName.BATTERY_SOC_RANGE_START, 2022),
                IntField(FieldName.BATTERY_SOC_RANGE_END, 2023),
            ],
        )



import asyncio
import logging
from typing import Any, Callable, List

from ..fields import FieldName
from ..base_devices import BluettiDevice, BaseDeviceV1, BaseDeviceV2
from ..bluetooth import DeviceReader, DeviceReaderConfig

_LOGGER = logging.getLogger(__name__)


class DeviceRecognizerResult:
    def __init__(
        self, name: str, iot_version: int, encrypted: bool, sn: int | None = None
    ):
        self.name = name
        self.iot_version = iot_version
        self.encrypted = encrypted
        self.sn = sn
        self.full_name = name + str(sn)


async def recognize_device(
    mac: str,
    future_builder_method: Callable[[], asyncio.Future[Any]],
) -> DeviceRecognizerResult | None:
    # Since we don't know the type we use the base device
    bluetti_devices: List[BluettiDevice] = [
        BaseDeviceV2(),
        BaseDeviceV1(),
    ]

    for bluetti_device in bluetti_devices:
        # Create device builder
        device_readers = [
            DeviceReader(
                mac,
                bluetti_device,
                future_builder_method,
                DeviceReaderConfig(
                    timeout=15,
                    use_encryption=True,
                ),
            ),
            DeviceReader(
                mac,
                bluetti_device,
                future_builder_method,
                DeviceReaderConfig(timeout=8),
            ),
        ]

        for device_reader in device_readers:

            # We only need 6 registers to get the device type
            data = await device_reader.read(
                bluetti_device.get_device_type_registers(),
            )

            if data is None:
                continue

            type_data = data.get(FieldName.DEVICE_TYPE.value)

            if type_data is None:
                # We have a problem
                _LOGGER.error("No data in device type type_data")
                continue

            if not isinstance(type_data, str):
                # We have a problem
                _LOGGER.error("Invalid data in device type type_data")
                continue

            if type_data == "":
                # Empty string is not a valid device type
                continue

            data = await device_reader.read(
                bluetti_device.get_device_sn_registers(),
            )

            if data is None:
                # Should never happen
                return DeviceRecognizerResult(
                    type_data,
                    bluetti_device.get_iot_version(),
                    device_reader.config.use_encryption,
                    "000000000000",  # Use dummy SN
                )

            sn_data = data.get(FieldName.DEVICE_SN.value)

            if not isinstance(sn_data, int) or sn_data == "":
                # Should never happen
                return DeviceRecognizerResult(
                    type_data,
                    bluetti_device.get_iot_version(),
                    device_reader.config.use_encryption,
                    "000000000000",  # Use dummy SN
                )

            return DeviceRecognizerResult(
                type_data,
                bluetti_device.get_iot_version(),
                device_reader.config.use_encryption,
                sn_data,
            )

    return None



"""Device builder helper."""

from ..base_devices import BluettiDevice

from ..devices import DEVICES, DEVICE_NAME_RE


def build_device(name: str) -> BluettiDevice | None:
    devMatch = DEVICE_NAME_RE.match(name)

    if devMatch is None:
        return None

    devType = devMatch[1]

    if devType is None:
        return None

    Station = DEVICES.get(devType)

    if Station is None:
        return None

    return Station()



"""Bluetti detect command."""

import argparse
import asyncio
import logging

from ..bluetooth.device_recognizer import recognize_device


async def async_detect_device(address: str):
    print("Detecting device type")
    print()

    recognized = await recognize_device(address, asyncio.Future)

    if recognized is None:
        print("Unable to find device type information")
        return

    print()
    print(
        f"Device type is '{recognized.name}' with iot version {recognized.iot_version} and serial {recognized.sn}. Full name: {recognized.full_name}"
    )
    if recognized.encrypted:
        print("This device uses encryption.")


def start():
    """Entrypoint."""
    parser = argparse.ArgumentParser(description="Detect bluetti devices")
    parser.add_argument("mac", type=str, help="Mac-address of the powerstation")
    args = parser.parse_args()

    if args.mac is None:
        parser.print_help()
        return

    # logging.basicConfig(level=logging.DEBUG)

    asyncio.run(async_detect_device(args.mac))



"""Setup for pypi package"""

import os
import codecs
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = os.getenv("LIB_VERSION")
DESCRIPTION = "Bluetti BT"

# Setting up
setup(
    name="bluetti-bt-lib",
    version=VERSION,
    author="Patrick762",
    author_email="<pip-bluetti-bt-lib@hosting-rt.de>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/Patrick762/bluetti-bt-lib",
    packages=find_packages(),
    install_requires=[
        "async_timeout",
        "asyncio",
        "bleak",
        "bleak_retry_connector",
        "crcmod",
        "cryptography",
        "logging",
        "pyasn1",
    ],
    keywords=[],
    entry_points={
        "console_scripts": [
            "bluetti-scan = bluetti_bt_lib.scripts.bluetti_scan:start",
            "bluetti-detect = bluetti_bt_lib.scripts.bluetti_detect:start",
            "bluetti-read = bluetti_bt_lib.scripts.bluetti_read:start",
            "bluetti-readall = bluetti_bt_lib.scripts.bluetti_readall:start",
            "bluetti-write = bluetti_bt_lib.scripts.bluetti_write:start",
            "bluetti-parse = bluetti_bt_lib.scripts.bluetti_parse:start",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
    ],
)



from typing import List

from . import BluettiDevice
from ..fields import DeviceField
from ..fields import FieldName, StringField, UIntField, SerialNumberField
from ..registers import ReadableRegisters


class BaseDeviceV1(BluettiDevice):
    def __init__(self, additional_fields: List[DeviceField] = []):
        super().__init__(
            [
                StringField(FieldName.DEVICE_TYPE, 10, 6),
                SerialNumberField(FieldName.DEVICE_SN, 17),
                UIntField(FieldName.BATTERY_SOC, 43, min=0, max=100),
                UIntField(FieldName.DC_INPUT_POWER, 36),
                UIntField(FieldName.AC_INPUT_POWER, 37),
                UIntField(FieldName.AC_OUTPUT_POWER, 38),
                UIntField(FieldName.DC_OUTPUT_POWER, 39),
            ]
            + additional_fields,
        )

    def get_full_registers_range(self) -> List[ReadableRegisters]:
        return [ReadableRegisters(i, 50) for i in range(1, 8000, 50)]

    def get_device_type_registers(self) -> List[ReadableRegisters]:
        return [
            ReadableRegisters(10, 6),
        ]

    def get_device_sn_registers(self) -> List[ReadableRegisters]:
        return [
            ReadableRegisters(17, 4),
        ]

    def get_iot_version(self) -> int:
        return 1



"""Bluetti BT Lib exports."""

from .base_devices import BluettiDevice
from .bluetooth import (
    DeviceReader,
    DeviceReaderConfig,
    DeviceWriter,
    DeviceRecognizerResult,
    recognize_device,
)
from .enums import *
from .fields import DeviceField, FieldName, FieldUnit, get_unit
from .utils.device_builder import build_device



from ..base_devices import BaseDeviceV2
from ..fields import FieldName, UIntField, SwitchField


class AC60(BaseDeviceV2):
    def __init__(self):
        super().__init__(
            [
                UIntField(FieldName.DC_OUTPUT_POWER, 140),
                UIntField(FieldName.AC_OUTPUT_POWER, 142),
                UIntField(FieldName.DC_INPUT_POWER, 144),
                UIntField(FieldName.AC_INPUT_POWER, 146),
                UIntField(FieldName.AC_INPUT_VOLTAGE, 1314),
                SwitchField(FieldName.CTRL_AC, 2011),
                SwitchField(FieldName.CTRL_DC, 2012),
                SwitchField(FieldName.CTRL_POWER_LIFTING, 2021),
            ],
        )



import struct

from . import DeviceField, FieldName


class BoolField(DeviceField):
    def __init__(self, name: FieldName, address: int):
        super().__init__(name, address, 1)

    def parse(self, data: bytes) -> bool:
        return struct.unpack("!H", data)[0] == 1



from . import DeviceField, FieldName


class StringField(DeviceField):
    def __init__(self, name: FieldName, address: int, size: int):
        super().__init__(name, address, size)

    def parse(self, data: bytes) -> str:
        return data.rstrip(b"\0").decode("ascii", errors="ignore")



import struct

from . import DeviceField, FieldName


class UIntField(DeviceField):
    def __init__(
        self,
        name: FieldName,
        address: int,
        multiplier: float = 1,
        min: int | None = None,
        max: int | None = None,
    ):
        super().__init__(name, address, 1)
        self.multiplier = multiplier
        self.min = min
        self.max = max

    def parse(self, data: bytes) -> int:
        val = struct.unpack("!H", data)[0]
        if self.multiplier != 1:
            val = round(val * self.multiplier, 2)
        return val

    def in_range(self, value: int) -> bool:
        if self.min is not None and self.min > value:
            return False
        if self.max is not None and self.max < value:
            return False
        return value >= 0



from ..base_devices import BaseDeviceV2
from ..fields import FieldName, UIntField, DecimalField, IntField


class EP760(BaseDeviceV2):
    def __init__(self):
        super().__init__(
            [
                UIntField(FieldName.PV_S1_POWER, 1212),
                DecimalField(FieldName.PV_S1_VOLTAGE, 1213, 1),
                DecimalField(FieldName.PV_S1_CURRENT, 1214, 1),
                UIntField(FieldName.PV_S2_POWER, 1220),
                DecimalField(FieldName.PV_S2_VOLTAGE, 1221, 1),
                DecimalField(FieldName.PV_S2_CURRENT, 1222, 1),
                UIntField(FieldName.SM_P1_POWER, 1228),
                DecimalField(FieldName.SM_P1_VOLTAGE, 1229, 1),
                UIntField(FieldName.SM_P1_CURRENT, 1230, 1),
                DecimalField(FieldName.GRID_FREQUENCY, 1300, 1),
                UIntField(FieldName.GRID_P1_POWER, 1313),
                DecimalField(FieldName.GRID_P1_VOLTAGE, 1314, 1),
                DecimalField(FieldName.GRID_P1_CURRENT, 1315, 1),
                DecimalField(FieldName.AC_OUTPUT_FREQUENCY, 1500, 1),
                IntField(FieldName.AC_P1_POWER, 1510),
                DecimalField(FieldName.AC_P1_VOLTAGE, 1511, 1),
                DecimalField(FieldName.AC_P1_CURRENT, 1512, 1),
            ],
        )



import struct
from enum import Enum
from typing import Any, Type, TypeVar

from . import DeviceField, FieldName


E = TypeVar("E", bound=Enum)


class SelectField(DeviceField):
    def __init__(self, name: FieldName, address: int, e: Type[E]):
        super().__init__(name, address, 1)
        self.e = e

    def parse(self, data: bytes) -> E | None:
        val = struct.unpack("!H", data)[0]
        return self.e(val)

    def is_writeable(self):
        return True

    def allowed_write_type(self, value: Any) -> bool:
        return isinstance(value, Type[Enum])
